/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: CsvProcessor.java 1089 2006-09-14 19:57:55Z dutoo $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.csv;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.apache.commons.io.IOUtils;
import org.objectweb.petals.component.common.listener.AbstractInternalMEProcessor;
import org.objectweb.petals.component.common.xmlmanipulation.SourceHelper;
import org.objectweb.petals.component.common.xmlmanipulation.StringHelper;
import org.objectweb.petals.engine.csv.exception.CsvException;

import com.csvreader.CsvReader;

/**
 * This processor uses Java CSV's CsvReader to transform either
 * the <content>-encapsulated content or the single attachment of
 * the message from any CSV format to XML.
 * 
 * It is configured through properties given by the CSV SU handler.
 * Here are their meaning :
 * <ul>
 *     <li>csvInputMode : may be "content", "attachment" or "guess".
 * Default is "guess", meaning that if there is any attachment this
 * will be the chosen mode, and if not, it will be in "content" mode.</li>
 *     <li>customColumnDelimiterChar : default is the comma ","</li>
 *     <li>useCustomRecordDelimiter : whether to use a custom one
 * instead of the default combination of unix & windows line breaks.
 * SO beware, using it lessens the original capabilities of the
 * parser and it is therefore not advised.</li>
 *     <li>customRecordDelimiterChar : custom line delimiter character.
 * See the "useCustomRecordDelimiter" property to activate this feature.</li>
 *     <li>useDoubleEscapeMode : whether to use doubled backslash to
 * trigger escaping instead of a single one.</li>
 *     <li>useComment : whether to parse and make comments available.
 * Default is false.</li>
 *     <li>commentChar : the character used to starts comments.
 * Default is sharp "#".</li>
 *     
 * @version $Rev: 1081 $Date: {date}
 * @since   Petals 1.0
 * @author Marc DUTOO - Open Wide
 *
 */
public class CsvProcessor extends AbstractInternalMEProcessor {

    public static final String CSV_INPUT_MODE_CONTENT = "content";
    public static final String CSV_INPUT_MODE_ATTACHMENT = "attachment";
    public static final String CSV_INPUT_MODE_GUESS = "guess";

    public static final String CSV_INPUT_MODE_PROP = "csvInputMode";
    public static final String CUSTOM_COLUMN_DELIMITER_CHAR_PROP = "customColumnDelimiterChar";
    public static final String USE_CUSTOM_RECORD_DELIMITER_PROP = "useCustomRecordDelimiter";
    public static final String CUSTOM_RECORD_DELIMITER_CHAR_PROP = "customRecordDelimiterChar";
    public static final String USE_DOUBLE_ESCAPE_MODE_PROP = "useDoubleEscapeMode";
    public static final String USE_COMMENT_PROP = "useComment";
    public static final String COMMENT_CHAR_PROP = "commentChar";

    
    /** service name to component properties map */
    private Map<String, Properties> mapEndpointCsv;
    
    
    /**
     * Creates a new CSV processor configured to use the given map of
     * service names to component properties.
     * @param c
     * @param l
     * @param mapEndpointCsv
     */
    public CsvProcessor(DeliveryChannel c, Logger l,
        Map<String, Properties> mapEndpointCsv) {
        super(c, l);
        this.mapEndpointCsv = mapEndpointCsv;
    }


    @Override
    protected void sendInMessage(QName service, QName operation,
        NormalizedMessage in) throws CsvException {
        // not supported : since it is a service engine, it outputs something
        throw new CsvException("Only InOut messages are supported");
    }

    @Override
    protected boolean sendInOutMessage(QName service, QName operation,
        NormalizedMessage in, NormalizedMessage out, boolean optionalOut)
    throws Exception {
        // getting the service properties
        Properties csvProperties = getCsvProperties(service.getLocalPart());
        
        // getting the content to be transformed
        String csvContent = getCsvInputContent(in, csvProperties);

        if (StringHelper.isEmpty(csvContent)) {
            l.log(Level.WARNING, "");
            // this will be automaticly added to a Fault if the message
            throw new CsvException("operation : message empty");
        }

        // actual CSV to XML transformation
        ByteArrayOutputStream xmlOut = new ByteArrayOutputStream();
        try{
            transformCsvToXml(new ByteArrayInputStream(csvContent.getBytes()),
                xmlOut, csvProperties);
        } catch(Exception e) {
            String msg = "Error while transforming csv to XML : " + e.getMessage();
            l.log(Level.SEVERE, msg, e);
            throw new CsvException(msg, e);
        }

        // Transform StreamResult to String , to be send in NormalizeMessage
        String str= new String(xmlOut.toByteArray());
        if (str != null && !("".equals(str))) {
            out.setContent(SourceHelper.createSource(str));
        } else {
            throw new CsvException("Empty result message : Error in CSV to XML transformation");
        }
        return true;
    }


    /**
     * Returns the actual CSV content from the given incoming message
     * according to the processor's settings
     * @param in
     * @param csvProperties
     * @return
     * @throws Exception
     */
    protected String getCsvInputContent(NormalizedMessage in,
            Properties csvProperties) throws Exception {
        String contentString;
        if (isModeAttachment(in, csvProperties)) {
            // getting first attachment
            String firstAttachmentName = null;
            for (Iterator nameIt = in.getAttachmentNames().iterator(); nameIt.hasNext();) {
                firstAttachmentName = (String) nameIt.next();
                break;
            }
            if (firstAttachmentName == null) {
                return null; // no content
            }
            DataHandler firstAttachment = in.getAttachment(firstAttachmentName);
            contentString = IOUtils.toString(firstAttachment.getInputStream());

        } else { // inlined content mode
            Source content = in.getContent();
            contentString = createStringFromContent(content);
        }
        return contentString;
    }
    
    /**
     * Unwrapps the given content source : transforms the source
     * to a String and removes the (required) encapsulating <content>
     * tag.
     * @param content
     * @return
     * @throws Exception
     */
    protected static String createStringFromContent(Source content) throws Exception {
        //TODO with CDATA, handle special cases ??
        String contentString = SourceHelper.createString(content);
        if (!contentString.startsWith("<content>")) {
            throw new Exception("Expected <content> encapsulated content "
                + "but found " + contentString);
        }
        // Usual case : the original content has been encapsulated as XML
        // NB. since contentString results from an XSL transformation,
        // we know that it ends with </content>
        return contentString.substring(9, contentString.length() - 10);
    }

    /**
     * Returns whether the CSV is to be looked for in the
     * message's first attachment instead of in its content
     * @param in
     * @param csvProperties
     * @return
     */
    protected boolean isModeAttachment(NormalizedMessage in,
            Properties csvProperties) {
        String csvInputMode = csvProperties.getProperty(CSV_INPUT_MODE_PROP, "guess");
        if (CSV_INPUT_MODE_ATTACHMENT.equals(csvInputMode)) {
            return true;
        } else if (CSV_INPUT_MODE_CONTENT.equals(csvInputMode)) {
            return false;
        } else { // guessing mode
            if (!in.getAttachmentNames().isEmpty()) {
                return true;
            }
        }
        return false;
    }


    /**
     * Transforms the given CSV content input in XML and
     * outputs it
     * @param csvIn
     * @param xmlOut where the transformed CSV XML is written
     * @param csvProperties used to configure CSV parsing
     * @throws CsvException
     */
    protected void transformCsvToXml(InputStream csvIn, OutputStream xmlOut,
            Properties csvProperties) throws CsvException {
        if (csvIn == null) {
            throw new CsvException("Input stream must not be null");
        }
        // creating a new CsvReader :
        CsvReader csvReader = createCsvReader(csvIn, csvProperties);
        if (csvReader == null) {
            throw new CsvException("not a CSV file");
        }

        try {
            // now creating and outputting the transformed XML :
            StringBuffer xmlBuf = new StringBuffer();
            xmlBuf.append("<data>");
            while (csvReader.readRecord()) {
                xmlBuf.append("<row>");
                int columnNb = csvReader.getColumnCount();
                for (int i = 0; i < columnNb; i++) {
                    String value = csvReader.get(i);
                    value = (value == null) ? "" : value; //TODO required ??
                    xmlBuf.append("<cell>");
                    xmlBuf.append(value);
                    xmlBuf.append("</cell>");
                }
                xmlBuf.append("</row>");
            }
            xmlBuf.append("</data>");
            xmlOut.write(xmlBuf.toString().getBytes());
        } catch (Exception e) {
            throw new CsvException("Error while parsing CSV : " + e.getMessage());
        }
    }


    /**
     * Creates a new CsvReader configured according to the
     * component's properties, ready to parse CSV content.
     * @param csvIn the CSV content to be parser
     * @param csvProperties used to configure the CsvReader
     * @return a new, configured CsvReader
     * @throws CsvException if creation or configuration error
     */
    protected CsvReader createCsvReader(InputStream csvIn,
            Properties csvProperties) throws CsvException {
        try {
            // TODO use BeanUtils to set the properties ??
            char csvColumnDelimiterChar = csvProperties
                  .getProperty(CUSTOM_COLUMN_DELIMITER_CHAR_PROP, ",").charAt(0);
            char csvCustomRecordDelimiterChar = csvProperties
                  .getProperty(CUSTOM_RECORD_DELIMITER_CHAR_PROP, "\\n").charAt(0);
            boolean useCustomRecordDelimiter = Boolean.valueOf(csvProperties
                .getProperty(USE_CUSTOM_RECORD_DELIMITER_PROP, "false")).booleanValue();
            int escapeMode = (Boolean.valueOf(csvProperties.getProperty(
                USE_DOUBLE_ESCAPE_MODE_PROP, "true")).booleanValue()) ?
                CsvReader.ESCAPE_MODE_DOUBLED : CsvReader.ESCAPE_MODE_BACKSLASH;
            boolean useComment = Boolean.valueOf(csvProperties
                .getProperty(USE_COMMENT_PROP, "false")).booleanValue();
            char commentChar = csvProperties.getProperty(COMMENT_CHAR_PROP, "#").charAt(0);
        
            // creating CsvReader for given input
            // csvCharset : Csv SE default is inited in TODO
            CsvReader csvReader = new CsvReader(csvIn, Charset.defaultCharset());
            csvReader.setDelimiter(csvColumnDelimiterChar); // CsvReader default is ','
            if (useCustomRecordDelimiter) {
               csvReader.setRecordDelimiter(csvCustomRecordDelimiterChar);
            } // else CsvReader default is a combination of unix & windows cr & lf
            csvReader.setEscapeMode(escapeMode); // CsvReader default is CsvReader.ESCAPE_MODE_DOUBLED i.e. '\\'
            csvReader.setUseComments(useComment);
            csvReader.setComment(commentChar); // CsvReader default is '#'
            // NB. safetySwitch is activated and may not be disabled for now
            //csvReader.setMergeConsecutive(arg0) //TODO ???
            return csvReader;
        } catch (Exception e) {
            throw new CsvException("Error setting CSV reader configuration."
                + "CSV reader properties are : " + csvProperties);
        }
    }
    
    
    /**
     * Returns the component configuration (for CsvReader etc.)
     * corresponding to the given service name
     * @param serviceName
     * @return
     * @throws Exception
     */
    protected Properties getCsvProperties(String serviceName) throws Exception{
        Properties csvProperties = this.mapEndpointCsv.get(serviceName);
        if (csvProperties != null) {
            return csvProperties;
        } else {
            throw new Exception("CSV properties not found for service "
                + serviceName);
        }
    }

}
