/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JmsListeningSession.java 12:18:02 alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.jms.incoming;

import java.util.Enumeration;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.jms.Connection;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageListener;
import javax.jms.QueueConnection;
import javax.jms.QueueSession;
import javax.jms.Session;
import javax.jms.TopicConnection;
import javax.jms.TopicSession;
import javax.jms.TextMessage;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import org.objectweb.petals.binding.jms.JmsConnection;
import org.objectweb.petals.component.common.xmlmanipulation.SourceHelper;

/**
 * This class is the JMS message listener for the JmsBC component
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.0
 * @author alouis
 */
public class JmsListeningSession implements MessageListener {

    public static final String PROPERTY_OPERATION = "javax.jbi.messaging.operation";

    protected ComponentContext ctx;

    protected DeliveryChannel channel;

    protected Logger l;

    protected JmsConnection jmsConnection;

    protected JbiConnection jbiConnection;

    protected MessageExchangeFactory exchangeFactory;

    protected Connection connection;

    protected Session session;

    protected MessageConsumer messageConsumer;

    public JmsListeningSession(ComponentContext c, DeliveryChannel channel,
        JmsConnection jmsConnection, JbiConnection jbiConnection, Logger log) {
        ctx = c;

        l = log;

        this.channel = channel;

        this.jmsConnection = jmsConnection;

        this.jbiConnection = jbiConnection;
    }

    public void init() throws Exception {
        jmsConnection.init();
    }

    public void listen() throws JMSException {
        if (jmsConnection.isTopicConnection()) {
            listenTopic();
        } else if (jmsConnection.isQueueConnection()) {
            listenQueue();
        } else {

        }
    }

    /**
     * Create a subscriber and 
     * Listen on the topic specified in the Jmsconnection object
     * @throws JMSException
     */
    protected void listenTopic() throws JMSException {
        l.log(Level.FINE, "listenTopic");

        // create a subscriber
        TopicConnection tc = jmsConnection.getTcf().createTopicConnection();
        connection = tc;
        TopicSession ts = tc.createTopicSession(true, 0);
        session = ts;
        messageConsumer = ts.createSubscriber(jmsConnection.getTopic());

        // set This as the listener
        messageConsumer.setMessageListener(this);

        // start listening
        tc.start();
    }

    /**
     * Create a receiver and 
     * Listen on the queue specified in the Jmsconnection object
     * @throws JMSException
     */
    protected void listenQueue() throws JMSException {
        l.log(Level.FINE, "listenQueue");

        // create a receiver
        QueueConnection qc = jmsConnection.getQcf().createQueueConnection();
        connection = qc;
        QueueSession qs = qc.createQueueSession(true, 0);
        session = qs;
        messageConsumer = qs.createReceiver(jmsConnection.getQueue());

        // set This as the listener
        messageConsumer.setMessageListener(this);

        //start listening
        qc.start();
    }

    public void terminate() throws JMSException {
        l.log(Level.FINE, "terminate");

        messageConsumer.close();
        session.close();
        connection.close();
    }

    public void onMessage(Message jmsExchange) {
        l.log(Level.FINE, "onMessage");
        l.log(Level.INFO, "A JMS message is received from '"
            + jmsConnection.getDestinationName()
            + "'.\nSend it to the JBI Endpoint '"
            + jbiConnection.getEndpointName() + "'");

        try {
            // 1. Only TextMessage are read.
            TextMessage textMessage;
            textMessage = (TextMessage) jmsExchange;

            // 2. Create an exchange
            InOnly exchange = createInOnly();

            // 3. transform textMessage->normalizedMessage
            transform(textMessage, exchange.getInMessage());

            // 4. set optionnal operation
            setOperation(textMessage, exchange);

            // 5. send the exchange
            channel.send(exchange);
            l.log(Level.INFO, "Message sent to '"
                + jbiConnection.getEndpointName() + "'.");

        } catch (ClassCastException e) {
            l
                .log(
                    Level.SEVERE,
                    "Only TextMessage are processed by petals-binding-jms component.",
                    e);
        } catch (Exception e) {
            l.log(Level.SEVERE, e.getClass() + " " + e.getMessage(), e);
        }
    }

    /**
     * Create an InOnly exchange,
     * with Endpoint and/or InterfaceName set as defined
     * in the JbiConnection variable.
     * @return
     * @throws MessagingException
     */
    protected InOnly createInOnly() throws MessagingException {
        MessageExchangeFactory factory = getMessageExchangeFactory();
        InOnly exchange = factory.createInOnlyExchange();
        NormalizedMessage nm = exchange.createMessage();
        exchange.setInMessage(nm);

        // set address : endpoint or interface
        ServiceEndpoint endpoint = ctx.getEndpoint(jbiConnection
            .getServiceName(), jbiConnection.getEndpointName());
        exchange.setEndpoint(endpoint);
        exchange.setInterfaceName(jbiConnection.getInterfaceName());

        return exchange;
    }

    /**
     * 
     * @see org.objectweb.petals.component.common.xmlmanipulation.SourceHelper#createSource(java.lang.Source)
     */
    protected Source createSource(String input) {
        return SourceHelper.createSource(input);
    }

    /**
     * Return an unique MessageExchangeFactory created by the
     *  DeliveryChannel. Lazyloading pattern.
     * @return
     */
    protected MessageExchangeFactory getMessageExchangeFactory() {
        if (exchangeFactory == null) {
            exchangeFactory = channel.createExchangeFactory();
        }
        return exchangeFactory;
    }

    /**
     * Transform the TextMessage
     * into a NormalizedMessage of a JBI MsgExch .
     * The content and properties are set. 
     * @param jms
     * @param jbi
     * @throws JMSException getting jms content or properties
     * @throws MessagingException setting NormalizedMessage properties or content
     */
    protected void transform(TextMessage jms, NormalizedMessage jbi)
        throws JMSException, MessagingException {
        l.log(Level.FINE, "transform");

        //set content
        String content = jms.getText();
        Source source = SourceHelper.createSource(content);

        jbi.setContent(source);

        // set properties
        for (Enumeration it = jms.getPropertyNames(); it.hasMoreElements();) {
            String pName = (String) it.nextElement();
            jbi.setProperty(pName, jms.getObjectProperty(pName));
        }
    }

    /**
     * Set the operation attribute of the exchange,
     * if an 'operation' property is set on the jms message
     * @param jms
     * @param jbi
     * @throws JMSException getting jms properties
     */
    protected void setOperation(TextMessage jms, MessageExchange jbi)
        throws JMSException {
        l.log(Level.FINE, "setOperation");

        String operation = jms.getStringProperty(PROPERTY_OPERATION);

        if (operation != null)
            jbi.setOperation(new QName(operation));
    }
}
