/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: EvalRateListener.java 10:42:15 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.demo.mortgage.evalrate;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;

import org.objectweb.petals.component.common.xmlmanipulation.SourceHelper;
import org.objectweb.petals.component.common.xmlmanipulation.StringHelper;

/**
 * Listener for the evaluation rate service engine of the Mortgage demonstration
 * 
 * @author ddesjardins - eBMWebsourcing
 * 
 */
public class EvalRateListener implements Runnable {

    private DeliveryChannel channel;

    private Logger logger;

    private MessageExchange messageExchange;

    private boolean running;

    public EvalRateListener(DeliveryChannel channel, Logger logger) {
        this.channel = channel;
        this.logger = logger;
    }

    public void run() {
        running = true;
        while (running) {
            try {
                messageExchange = channel.accept();
                if (messageExchange != null) {
                    process(messageExchange);
                }
            } catch (MessagingException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }
    }

    public void stopProcessing() {
        this.running = false;
    }

    private void process(MessageExchange msg) {
        logger.log(Level.INFO, "");
        try {
            /*
             * receive an acquittement ?
             */
            if (msg.getStatus().equals(ExchangeStatus.DONE)) {
                logger.log(Level.INFO, "receive a status DONE. nothing to do.");
                return;
            }

            // Only "evaluate" method and InOut Mep are allowed
            if (new QName("evaluate").equals(msg.getOperation())
                    && msg instanceof InOut) {
                doEvaluate(msg);
            } else {
                Exception e = new Exception(
                        "only 'evaluate/InOut operation is recognized");
                msg.setError(e);
                msg.setStatus(ExchangeStatus.ERROR);
                logger.log(Level.WARNING, e.getMessage(), e);
            }
            channel.send(msg);
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    /**
     * Execute the "evaluate" operation : evaluate a mortgage rate based on the
     * given profile. Two type of mortgage rate evaluations are processed by
     * this method : simple (only "yes" or "no" profile allowed) and detailed
     * (good, medium and risky profile allowed).
     * 
     * @param msg
     *            The message exchange that handles profile information
     * @throws Exception
     */
    protected void doEvaluate(MessageExchange msg) throws Exception {
        /*
         * Read the IN message
         */
        String content = SourceHelper.createString(msg.getMessage("IN")
                .getContent());
        String response = "";

        // Content must not be empty
        if (StringHelper.isEmpty(content)) {
            Fault f = msg.createFault();
            response = "You must specify the profil of the mortgage subscriber";
            logger.log(Level.WARNING, "The EvalRate answers a fault: "
                    + response);
            f.setContent(SourceHelper.createSource(response));
            msg.setFault(f);
        } else {
            // Get the string content from the xml content
            content = content.substring(content.indexOf("<profile>")+9, content.indexOf("</profile>"));
            
            logger.info("Eval Rate Module called to evaluate mortgage rate based on the following profile : "
                    + content);
            // Only "yes" profile is allowed for a simple Mortgage
            // Rate evaluation (first Mortgage process)
            if ("Yes".equalsIgnoreCase(content)) {
                response = "5";
            }
            // For a Detailed Mortgage Rate evaluation, three different
            // profiles are allowed : Good, Medium and Risky
            else if ("Good".equalsIgnoreCase(content)) {
                response = "4";
            } else if ("Medium".equalsIgnoreCase(content)) {
                response = "5";
            } else if ("Risky".equalsIgnoreCase(content)) {
                response = "6";
            }

            // Other profile are not allowed
            if ("".equals(response)) {
                Fault f = msg.createFault();
                response = "Given mortgage profile isn't allowed: " + content;
                logger.log(Level.WARNING, "The EvalRate answers a fault: "
                        + response);
                f.setContent(SourceHelper.createSource(response));
                msg.setFault(f);
            } else {
                logger.log(Level.INFO, "The EvalRate answer : " + response);
                
                // Create xml content from string content
                response = "<response>"+response+"</response>";
                
                NormalizedMessage nm = msg.createMessage();
                nm.setContent(SourceHelper.createSource(response));
                msg.setMessage(nm, "OUT");
            }
        }
    }

}
