/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: MortgageWorkflowProcessor.java 154 6 juin 2006 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.demo.mortgage.workflow;

import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.objectweb.petals.component.common.listener.IMessageExchangeProcessor;
import org.objectweb.petals.component.common.xmlmanipulation.SourceHelper;
import org.objectweb.petals.demo.mortgage.workflow.orchestration.DetailedMortgageProfiling;
import org.objectweb.petals.demo.mortgage.workflow.orchestration.MortgageProfiling;
import org.objectweb.petals.demo.mortgage.workflow.orchestration.MortgageRequest;
import org.objectweb.petals.demo.mortgage.workflow.orchestration.MortgageResponse;
import org.objectweb.petals.demo.mortgage.workflow.orchestration.SimpleMortgageProfiling;

/**
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class MortgageWorkflowProcessor implements IMessageExchangeProcessor {

    protected ComponentContext ctx;
    
    protected DeliveryChannel channel;

    protected Logger l;

    public MortgageWorkflowProcessor(ComponentContext ctx,
            DeliveryChannel channel, Logger l) {
        super();
        this.ctx = ctx;
        this.channel = channel;
        this.l = l;
    }

    public boolean process(MessageExchange exchange) throws Exception {

        if (exchange == null || !(exchange instanceof InOut)) {
            throw new Exception("Mortgage workflow accept in out exchange only");
        }

        if (ExchangeStatus.DONE.equals(exchange.getStatus())) {
            l
                    .info("Mortgage Workflow service receives a status DONE. Nothing to do !");
            // Do nothing
        } else if (new QName("orchestrate").equals(exchange.getOperation())) {

            /*
             * Start workflow
             */
            orchestrate(exchange);

        } else {
            throw new NoSuchMethodException("Only start method supported");
        }

        return true;
    }

    /**
     * Start the message processing : BO creation and execution of the workflow
     * 
     * @param exchange
     *            the MessageExchange that handles the mortgage request and
     *            response
     * @throws Exception
     */
    private void orchestrate(MessageExchange exchange) throws Exception {

        /*
         * Retrieve the endpointName
         */
        String endpointName = "";
        ServiceEndpoint endpoint = exchange.getEndpoint();
        if (endpoint == null) {
            throw new Exception("Exchange must have an endpoint");
        } else {
            endpointName = endpoint.getEndpointName();
        }

        /*
         * Retrieve message content
         */
        NormalizedMessage inMess = ((InOut) exchange).getInMessage();
        Source content = inMess.getContent();
        if (content == null) {
            throw new Exception("InMessage content must be non null");
        }

        /*
         * retrieve request parameters
         */
        MortgageRequest request = retrieveRequestParameters(content);

        /*
         * execute workflow to process
         */
        MortgageResponse response = executeWorkflow(endpointName, request);
        
        /*
         * Send the response to the gui
         */
        sendResponse(response, exchange);
    }

    /**
     * Create the outMessage from the MortgageResponse and send it.
     * 
     * @param response
     *            the MortgageResponse bo
     * @param exchange
     *            the MessageExchange that handles the outMessage
     * @throws Exception
     */
    protected void sendResponse(MortgageResponse response,
            MessageExchange exchange) throws Exception {
        /*
         * Create Content from MortgageResponse
         */
        Source content = createContentFromMortgageResponse(response);

        /*
         * Create the outMessage to send
         */
        NormalizedMessage outNM = exchange.createMessage();
        outNM.setContent(content);
        ((InOut) exchange).setOutMessage(outNM);

        /*
         * Send the message exchange
         */
        this.channel.send(exchange);

    }

    /**
     * Create a Source content from the MortgageResponse. This Source handles a
     * Soap representation of the MortgageResponse bo.
     * 
     * @param response
     *            the MortgageResponse bo.
     * @return the Source handling the Soap representation of the
     *         MortgageResponse. Can't be null.
     */
    protected Source createContentFromMortgageResponse(MortgageResponse response) {

        StringBuffer soap = new StringBuffer();
        soap
                .append("<orchestrateResponse xmlns:soapenc=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\">");
        soap
                .append("<orchestrateReturn xmlns:tns0=\"http://orchestration.workflow.mortgage.demo.petals.objectweb.org\" xsi:type=\"tns0:MortgageResponse\">");
        soap.append("<firstName xsi:type=\"xsd:string\">"
                + response.getFirstName() + "</firstName>");
        soap.append("<lastName xsi:type=\"xsd:string\">"
                + response.getLastName() + "</lastName>");
        soap.append("<mortgageRate xsi:type=\"xsd:float\">"
                + response.getMortgageRate() + "</mortgageRate>");
        soap.append("</orchestrateReturn>");
        soap.append("</orchestrateResponse>");

        return SourceHelper.createSource(soap.toString());
    }

    /**
     * Execute the workflow depending on the endpoint information : detailed
     * workflow for "detailedMortgageProfiler" endpoint and simple workflow for
     * "simpleMortgageProfiler" endpoint.
     * 
     * @param endpointName
     *            name of the targeted endpoint
     * @param request
     *            MortgageRequest bo
     * @return MortgageResponse bo. Can't be null.
     * @throws Exception
     *             if given endpoint doesn't exist for MortgageWorkflow service
     */
    private MortgageResponse executeWorkflow(String endpointName,
            MortgageRequest request) throws Exception {
        MortgageResponse response = null;

        MortgageProfiling profiling = null;
        if (MortgageWorkflow.ENDPOINT1.equals(endpointName)) {
            profiling = new SimpleMortgageProfiling(ctx, channel, l);
        } else if (MortgageWorkflow.ENDPOINT2.equals(endpointName)) {
            profiling = new DetailedMortgageProfiling(ctx, channel, l);
        } else {
            throw new Exception(
                    "Given endpoint doesn't exist for MortgageWorkflow");
        }
        response = profiling.orchestrate(request);

        return response;
    }

    /**
     * Build a MortgageRequest bo from information contained into the Source
     * content (a Soap representation of the MortgageRequest).
     * 
     * @param content
     *            the Source content (a Soap representation of the
     *            MortgageRequest)
     * @return a MortgageRequest bo. Can't be null.
     * @throws Exception
     */
    protected MortgageRequest retrieveRequestParameters(Source content)
        throws Exception {

        String stringContent = SourceHelper.createString(content);

        String address = "";
        String autoPayment = "";
        String creditCards = "";
        String firstName = "";
        String insurance = "";
        String lastName = "";
        String otherPayments = "";
        String propertyTaxes = "";
        String salary = "";

        if (stringContent.indexOf("<address") > -1) {
            address = stringContent.substring(
                    stringContent.indexOf("<address") + 31, stringContent
                            .indexOf("</address>"));
        }
        if (stringContent.indexOf("<autoPayment") > -1) {
            autoPayment = stringContent.substring(stringContent
                    .indexOf("<autoPayment") + 34, stringContent
                    .indexOf("</autoPayment>"));
        }
        if (stringContent.indexOf("<creditCards") > -1) {
            creditCards = stringContent.substring(stringContent
                    .indexOf("<creditCards") + 34, stringContent
                    .indexOf("</creditCards>"));
        }
        if (stringContent.indexOf("<firstName") > -1) {
            firstName = stringContent.substring(stringContent
                    .indexOf("<firstName") + 33, stringContent
                    .indexOf("</firstName>"));
        }
        if (stringContent.indexOf("<insurance") > -1) {
            insurance = stringContent.substring(stringContent
                    .indexOf("<insurance") + 32, stringContent
                    .indexOf("</insurance>"));
        }
        if (stringContent.indexOf("<lastName") > -1) {
            lastName = stringContent.substring(stringContent
                    .indexOf("<lastName") + 32, stringContent
                    .indexOf("</lastName>"));
        }
        if (stringContent.indexOf("<otherPayments") > -1) {
            otherPayments = stringContent.substring(stringContent
                    .indexOf("<otherPayments") + 36, stringContent
                    .indexOf("</otherPayments>"));
        }
        if (stringContent.indexOf("<propertyTaxes") > -1) {
            propertyTaxes = stringContent.substring(stringContent
                    .indexOf("<propertyTaxes") + 36, stringContent
                    .indexOf("</propertyTaxes>"));
        }
        if (stringContent.indexOf("<salary") > -1) {
            salary = stringContent.substring(
                    stringContent.indexOf("<salary") + 29, stringContent
                            .indexOf("</salary>"));
        }

        MortgageRequest request = new MortgageRequest(firstName, lastName,
                address, new Float(salary).floatValue(), new Float(
                        propertyTaxes).floatValue(), new Float(insurance)
                        .floatValue(), new Float(autoPayment).floatValue(),
                new Float(creditCards).floatValue(), new Float(otherPayments)
                        .floatValue());

        return request;
    }

}
