/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: RouterImpl.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.routing;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.petals.jbi.component.ComponentContextImpl;
import org.objectweb.petals.jbi.management.lifecycle.ComponentLifeCycle;
import org.objectweb.petals.jbi.management.service.EndpointService;
import org.objectweb.petals.jbi.management.service.LifeCycleManagerService;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.jbi.registry.AbstractEndpoint;
import org.objectweb.petals.jbi.registry.ConsumerEndpoint;
import org.objectweb.petals.jbi.registry.InternalEndpoint;
import org.objectweb.petals.jbi.transport.TransportException;
import org.objectweb.petals.jbi.transport.Transporter;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;

import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_ENDPOINTSERVICE_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_LIFECYCLEMANAGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_TRANSPORTER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.MONOLOG_FACTORY_NAME;

/**
 * Routes messages to their destinations. Performs an itinerary resolution,
 * adding for example a transformatino service before delivering the message to
 * its recipient, and perform an address resolution, finding the endpoint
 * satisfying the the destination information (finding an endpoint for a
 * specified service). Resolutions are based on defined rules.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */
public class RouterImpl implements Router, BindingController,
    LifeCycleController {
    private static final String[] CLIENT_INTERFACES = new String[] {
        JBI_TRANSPORTER_NAME, JBI_LIFECYCLEMANAGER_NAME,
        JBI_ENDPOINTSERVICE_NAME};

    protected EndpointService endpointService;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    protected Logger logger;

    protected LifeCycleManagerService manager;

    protected Transporter transporter;

    protected AddressResolver addressResolver;

    /**
     * Component LifeCycle state.
     */
    private String fcState;

    /**
     * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
     *      java.lang.Object)
     */
    public void bindFc(String clientItfName, Object serverItf)
        throws NoSuchInterfaceException, IllegalBindingException,
        IllegalLifeCycleException {

        if (LOGGER_NAME.equals(clientItfName)) {
            logger = (Logger) serverItf;
        } else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
            // nothing
        } else if (JBI_ENDPOINTSERVICE_NAME.equals(clientItfName)) {
            endpointService = (EndpointService) serverItf;
        } else if (JBI_LIFECYCLEMANAGER_NAME.equals(clientItfName)) {
            manager = (LifeCycleManagerService) serverItf;
        } else if (JBI_TRANSPORTER_NAME.equals(clientItfName)) {
            transporter = (Transporter) serverItf;
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
     */
    public String getFcState() {
        return STARTED.equals(fcState) ? STARTED : STOPPED;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#listFc()
     */
    public String[] listFc() {
        return CLIENT_INTERFACES;
    }

    // ----------------------------------------------------------------------
    // Fractal BindingController implementation
    // ----------------------------------------------------------------------

    /**
     * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
     */
    public Object lookupFc(String clientItfName)
        throws NoSuchInterfaceException {

        if (JBI_LIFECYCLEMANAGER_NAME.equals(clientItfName)) {
            return manager;
        } else if (JBI_ENDPOINTSERVICE_NAME.equals(clientItfName)) {
            return endpointService;
        } else if (JBI_TRANSPORTER_NAME.equals(clientItfName)) {
            return transporter;
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    /**
     * Sends the specified message. Depending on the Role of the sender, send
     * the message to the producer (request) or to the consumer (response)
     * 
     * @param source
     * @param exchange
     * @param timeOut
     * @throws RoutingException
     */
    public void send(ComponentContextImpl source, MessageExchangeImpl exchange,
        long timeOut) throws RoutingException {
        log.call();

        if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())) {
            sendToProvider(source, exchange, timeOut);
        } else {
            sendToConsumer(source, exchange, timeOut);
        }
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    public void startFc() throws IllegalLifeCycleException {
        init();
        fcState = STARTED;
    }

    // ----------------------------------------------------------------------
    // Fractal LifecycleController implementation
    // ----------------------------------------------------------------------

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
     */
    public void stopFc() throws IllegalLifeCycleException {
        shutdown();
        fcState = STOPPED;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
     */
    public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
        IllegalBindingException, IllegalLifeCycleException {

        if (JBI_TRANSPORTER_NAME.equals(clientItfName)) {
            transporter = null;
        } else if (JBI_ENDPOINTSERVICE_NAME.equals(clientItfName)) {
            endpointService = null;
        } else if (JBI_LIFECYCLEMANAGER_NAME.equals(clientItfName)) {
            manager = null;
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    protected void init() {
        log = new LoggingUtil(logger);
        log.start();
        addressResolver = new AddressResolver(endpointService);
        log.end();
    }

    protected void shutdown() {
        log.start();

        logger = null;

        log.end();
    }

    /**
     * Sends the specified message, sent by the specified component. TODO send
     * directly to the DeliveryChannel of the component related to the recipient
     * endpoint.
     * 
     * @param source
     * @param exchange
     * @param timeOut
     * @throws RoutingException
     */
    protected void sendToConsumer(ComponentContextImpl source,
        MessageExchangeImpl exchange, long timeOut) throws RoutingException {
        log.start();

        try {
            AbstractEndpoint consumerEndpoint = exchange.getConsumerEndpoint();

            String containerName = consumerEndpoint.getContainerName();

            exchange.setRole(MessageExchange.Role.CONSUMER);

            transporter.send(exchange, containerName, timeOut);
        } catch (TransportException e) {
            // restore the Role if the message has not been sent
            exchange.setRole(MessageExchange.Role.PROVIDER);

            throw new RoutingException(e);
        }

        log.end();
    }

    /**
     * Sends the specified message to the provider of the service. TODO for the
     * moment, send directly to the DeliveryChannel of the component related to
     * the recipient endpoint.
     * 
     * @param source
     * @param exchange
     * @throws RoutingException
     */
    protected void sendToProvider(ComponentContextImpl source,
        MessageExchangeImpl exchange, long timeOut) throws RoutingException {
        log.start();

        // resolve the address (find endpoint)
        addressResolver.resolveAddress(source, exchange);

        try {
            AbstractEndpoint providerEndpoint = (AbstractEndpoint) exchange
                .getEndpoint();

            String containerName = providerEndpoint.getContainerName();

            exchange.setRole(MessageExchange.Role.PROVIDER);

            transporter.send(exchange, containerName, timeOut);
        } catch (TransportException e) {
            // restore the Role if the message has not been sent
            exchange.setRole(MessageExchange.Role.CONSUMER);

            // TODO set message to Error ?
            throw new RoutingException(e);
        }

        log.end();
    }

    /**
     * route the received exchange to the correct DeliveryChannel.
     */
    public void receive(MessageExchangeImpl exchange) throws RoutingException {
        log.start();

        // find the name of the component to deliver the Exchange.
        // Depends on the Role (consumer or provider)
        String componentname = null;

        if (MessageExchange.Role.PROVIDER.equals(exchange.getRole())) {
            componentname = ((InternalEndpoint) exchange.getEndpoint())
                .getComponentName();
        } else {
            componentname = ((ConsumerEndpoint) exchange.getConsumerEndpoint())
                .getComponentName();
        }

        // retrieve the ComponentLifeCycle
        ComponentLifeCycle comp = manager.getComponentByName(componentname);

        // push the exchange
        try {
            comp.getComponentContext().getDeliveryChannelImpl().push(exchange);
        } catch (JBIException e) {
            throw new RoutingException(e);
        }
        log.end();
    }
}
