/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: LoaderManagerImpl.java 1109 2006-09-19 09:29:19Z alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.classloader;

import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.MONOLOG_FACTORY_NAME;

import java.io.IOException;
import java.lang.reflect.Field;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.util.LoggingUtil;

/**
 * @version $Rev: 1109 $ $Date: 2006-09-19 11:29:19 +0200 (mar, 19 sep 2006) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>, Adrien
 *         LOUIS
 */
public class LoaderManagerImpl implements LoaderManager, BindingController {

    /**
     * Default constructor.
     */
    public LoaderManagerImpl() {
        super();
    }

    LoggingUtil logger;

    // ----------------------------------------------------------------------
    // ClassLoadingService interface implementation
    // ----------------------------------------------------------------------

    /**
     * @see org.objectweb.petals.classloader.LoaderManager#containsClassLoader(java.lang.String)
     */
    public boolean containsClassLoader(String classLoaderId) {
        logger.call();
        
        Loaders loaders = Loaders.getInstance();

        return loaders.contains(classLoaderId);
    }

    /**
     * Create a classLoader for a JBI Component. The parent classLoader will be
     * the SharedLibrariesClassLoader.
     * 
     * @see org.objectweb.petals.classloader.LoaderManager#createClassLoader(java.net.URL[],
     *      boolean)
     */
    public URLClassLoader createClassLoader(URL[] baseUrls,
        List<String> classpathLocations, boolean useParentFirst)
        throws PetalsException {

        logger.start();
        // create the class loader
        PetalsClassLoader cl = null;

        Loaders loaders = Loaders.getInstance();

        try {
            cl = new PetalsClassLoader(baseUrls, classpathLocations, loaders
                .getSharedLibrariesClassLoader());

            cl.setParentFirst(useParentFirst);
        } catch (IOException e) {
            throw new PetalsException("Error while creating a class loader", e);
        }
        logger.end();
        return cl;
    }

    /**
     * @see org.objectweb.petals.classloader.LoaderManager#createClassLoader(java.lang.String,
     *      java.net.URL[], boolean)
     */
    public URLClassLoader createComponentClassLoader(String classLoaderId,
        URL[] baseUrls, List<String> classpathUrls, boolean useParentFirst,
        List<String> sharedLibrariesNameList) throws PetalsException {
        
        logger.start();
        // create the class loader
        PetalsClassLoader cl = null;

        Loaders loaders = Loaders.getInstance();

        try {
            cl = new ComponentClassLoader(baseUrls, classpathUrls, loaders
                .getSharedLibrariesClassLoader(), sharedLibrariesNameList);

            cl.setParentFirst(useParentFirst);
        } catch (IOException e) {
            throw new PetalsException("Error while creating a class loader", e);
        }

        // save for further retrieve (overwrite existing class loader)
        loaders.put(classLoaderId, cl);

        logger.end();
        return cl;
    }

    /**
     * Create a classloader for a sharedLibrary. This classLoader will be added
     * to the SharedLibrariesClassLoader.
     * 
     * @see org.objectweb.petals.classloader.LoaderManager#createClassLoader(java.lang.String,
     *      java.net.URL[], boolean)
     */
    public URLClassLoader createSharedLibraryClassLoader(String classLoaderId,
        URL[] baseUrls, List<String> classpathUrls, boolean useParentFirst)
        throws PetalsException {

        logger.start();
        // create the class loader
        SharedLibraryClassLoader cl = null;

        try {
            cl = new SharedLibraryClassLoader(baseUrls, classpathUrls);
            cl.setParentFirst(useParentFirst);
        } catch (IOException e) {
            throw new PetalsException("Error while creating a class loader", e);
        }

        // save for further retrieve (overwrite existing class loader)
        Loaders loaders = Loaders.getInstance();

        loaders.put(classLoaderId, cl);
        loaders.getSharedLibrariesClassLoader().addSharedLibraryClassLoader(
            classLoaderId, cl);
        
        logger.end();
        return cl;
    }

    /**
     * Remove classLoader from the list. If the classLoader is a ClassLoader for
     * sharedLibraries, remove it from the SharedLibrariesClassLoader
     * 
     * @see org.objectweb.petals.classloader.LoaderManager#deleteClassLoader(java.lang.String)
     */
    public void deleteClassLoader(String classLoaderId) {

        logger.start();
        Loaders loaders = Loaders.getInstance();

        if (!containsClassLoader(classLoaderId)) {
            return;
        }
        // delete class loader for further retrieve
        PetalsClassLoader cLoader = loaders.remove(classLoaderId);

        cleanCache(cLoader);
        loaders.getSharedLibrariesClassLoader().removeSharedLibraryClassLoader(
            classLoaderId);
        cLoader = null;
        System.gc();
        logger.end();
    }

    // //////////////////////////////////////////////////////////////
    // clear class loader caches
    // ////////////////////////////////////////////////////////////////

    /**
     * When uninstalling a component, some elements remains in the cache of
     * certain classes, and the classloader can not release the JARS containing
     * those elements
     * 
     * referenced bug on Sun VM : 6232010
     * 
     */
    protected void cleanCache(PetalsClassLoader cLoader) {
        logger.call();
        
        System.gc();
        cleanCache(cLoader, "java.io.ObjectInputStream", "subclassAudits");
        cleanCache(cLoader, "java.io.ObjectOutputStream", "subclassAudits");
        cleanCache(cLoader, "java.io.ObjectStreamClass", "localDescs");
        cleanCache(cLoader, "java.io.ObjectStreamClass", "reflectors");
    }

    /**
     * Clear the cache contained in the specified class, for the specific field
     * (typically, streaming classes)
     * 
     * referenced bug on Sun VM : 6232010
     * 
     * @param the
     *            classLoader to clean (used when loading stream classes)
     * @param className
     *            the name of the class containing the cache field
     * @param fieldName
     *            the name of the cache field
     */
    protected void cleanCache(PetalsClassLoader cLoader, String className,
        String fieldName) {

        Map cache = null;
        try {
            Class clazz = cLoader.loadClass(className);
            for (Field field : clazz.getDeclaredFields()) {
                if (field.getType().equals(Map.class)
                    || field.getType().equals(HashMap.class)) {
                    field.setAccessible(true);
                    cache = (Map) field.get(null);
                    if (cache != null) {
                        synchronized (cache) {
                            cache.clear();
                        }
                    }
                }
            }
        } catch (Throwable ex) {
            // class not loaded, nothing to do
        }
    }
    
    

    // //////////////////////////////////////////////////////////////////
    // Fractal
    // //////////////////////////////////////////////////////////////////

    public void bindFc(String clientItfName, Object serverItf)
        throws NoSuchInterfaceException, IllegalBindingException,
        IllegalLifeCycleException {
        if (LOGGER_NAME.equals(clientItfName)) {
            logger = new LoggingUtil((Logger) serverItf);
        } else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
            // nothing
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    public String[] listFc() {
        return new String[0];
    }

    public Object lookupFc(String clientItfName)
        throws NoSuchInterfaceException {
        throw new NoSuchInterfaceException(clientItfName);
    }

    public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
        IllegalBindingException, IllegalLifeCycleException {
        if (LOGGER_NAME.equals(clientItfName)) {
            logger = null;
        } else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
            // nothing
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

}
