/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: RepositoryImpl.java 1115 2006-09-20 09:58:53Z ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.repository;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.Timer;
import java.util.TimerTask;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import org.apache.commons.io.FileUtils;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.petals.util.ZipUtil;
import org.objectweb.util.monolog.api.Logger;

import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.MONOLOG_FACTORY_NAME;

/**
 * Platform Component Repository service implementation.
 * 
 * @version $Rev: 1115 $ $Date: 2006-09-20 11:58:53 +0200 (mer, 20 sep 2006) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 * @author alouis - EBM Websourcing
 * @author ofabre - EBM Websourcing
 * @author ddesjardins - eBMWebsourcing
 */
public class RepositoryImpl implements RepositoryService, LifeCycleController,
        BindingController {

    /**
     * Path to component install directory
     */
    private static final String INSTALL_PATH = "install";

    /**
     * Path to repository location within Petals installation directory.
     */
    private static final String REPOSITORY_PATH = "repository";

    /**
     * Path to components directory into petals repository
     */
    private static final String COMPONENTS_PATH = "components";

    /**
     * Path to components directory into petals repository
     */
    private static final String SHAREDLIBS_PATH = "shared-libraries";

    /**
     * Path to lost+found directory into petals repository
     */
    private static final String LOSTPLUSFOUND_PATH = "lost+found";

    /**
     * Path to components directory into petals repository
     */
    private static final String SERVICEASSEMBLIES_PATH = "service-assemblies";

    /**
     * The complete repository directory path.
     */
    private static String repositoryDirectory;

    /**
     * Path to component working directory
     */
    private static final String WORK_PATH = "work";

    /**
     * Entity types (components, shared libs or service assemblies)
     */
    public static enum EntityType {
        COMPONENT_TYPE, SL_TYPE, SA_TYPE
    };

    /**
     * Component lifecycle state.
     */
    protected String fcState;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    protected Logger logger;

    /**
     * Delay between 2 deletion-attempts
     */
    private long DELETION_DELAY = 30;

    /**
     * number of deletion attempt
     */
    private int DELETION_ATTEMPT = 10;

    /**
     * @see org.objectweb.petals.repository.RepositoryService#addPackage(String,
     *      java.net.URI)
     */
    protected URI addPackage(String entityId, URI zipArchiveLocation,
            EntityType entityType) throws PetalsException {

        File installationRoot = null;

        // don't care about replacing existing data
        switch (entityType) {
        case COMPONENT_TYPE:
            if (containsComponentPackage(entityId)) {
                removeComponentPackage(entityId);
            }
            installationRoot = getComponentInstallRoot(entityId);
            break;
        case SA_TYPE:
            if (containsServiceAssemblyPackage(entityId)) {
                removeServiceAssemblyPackage(entityId);
            }
            installationRoot = getServiceAssemblyInstallRoot(entityId);
            break;
        case SL_TYPE:
            if (containsSharedLibPackage(entityId)) {
                removeSharedLibPackage(entityId);
            }
            installationRoot = getSharedLibInstallRoot(entityId);
            break;
        }
        File workRoot = new File(new File(installationRoot, ".."), WORK_PATH);
        
        // recreate dirs necessary for adding the package into the repository :
        // install
        if (!installationRoot.mkdirs()) {
            throw new PetalsException(
                    "Failed to create installation root dir: "
                            + installationRoot.getAbsolutePath());
        }
        // recreate dirs necessary for adding the package into the repository :
        // work
        if (!workRoot.mkdirs()) {
            throw new PetalsException("Failed to create work dir: "
                    + workRoot.getAbsolutePath());
        }

        // explode package Zip archive file into repository
        ZipFile zipArchive = null;

        try {
            File archiveFile = new File(zipArchiveLocation);
            zipArchive = new ZipFile(archiveFile);
        } catch (ZipException ze) {
            throw new PetalsException("Unable to open Zip archive file: "
                    + zipArchiveLocation.toString(), ze);
        } catch (IOException ioe) {
            throw new PetalsException("Unexpected exception.", ioe);
        }

        ZipUtil.explodeIntoDirectory(zipArchive, installationRoot);
        return installationRoot.toURI();

    }

    public URI addComponentPackage(String componentID, URI zipArchiveLocation)
        throws PetalsException {
        return addPackage(componentID, zipArchiveLocation,
                EntityType.COMPONENT_TYPE);
    }

    public URI addSharedLibPackage(String sharedLibID, URI zipArchiveLocation)
        throws PetalsException {
        return addPackage(sharedLibID, zipArchiveLocation, EntityType.SL_TYPE);
    }

    public URI addServiceAssemblyPackage(String serviceAssemblyID,
            URI zipArchiveLocation) throws PetalsException {
        return addPackage(serviceAssemblyID, zipArchiveLocation,
                EntityType.SA_TYPE);
    }

    public void bindFc(String clientItfName, Object serverItf)
        throws NoSuchInterfaceException, IllegalBindingException,
        IllegalLifeCycleException {
        if (LOGGER_NAME.equals(clientItfName)) {
            logger = (Logger) serverItf;
        } else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
            // nothing
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }

    }

    /**
     * @see org.objectweb.petals.repository.RepositoryService#containsPackage(java.lang.String)
     */
    protected boolean containsPackage(String entityID, EntityType entityType) {

        File entityRoot = null;

        switch (entityType) {
        case COMPONENT_TYPE:
            entityRoot = getComponentRoot(entityID);
            break;
        case SA_TYPE:
            entityRoot = getServiceAssemblyRoot(entityID);
            break;
        case SL_TYPE:
            entityRoot = getSharedLibRoot(entityID);
            break;
        }

        return entityRoot.exists() && entityRoot.isDirectory();
    }

    public boolean containsComponentPackage(String compId) {
        return containsPackage(compId, EntityType.COMPONENT_TYPE);
    }

    public boolean containsServiceAssemblyPackage(String saId) {
        return containsPackage(saId, EntityType.SA_TYPE);
    }

    public boolean containsSharedLibPackage(String slId) {
        return containsPackage(slId, EntityType.SL_TYPE);
    }

    /**
     * @see org.objectweb.petals.repository.RepositoryService#explodeSUIntoSAInstallFolder(String,
     *      URI, String)
     */
    public URI explodeSUIntoSAInstallFolder(String suId, URI suZipLocation,
            String saId) throws PetalsException {
        File installationRoot = new File(getServiceAssemblyInstallRoot(saId),
                suId);
        // recreate dirs necessary for adding the package into the repository
        // FIXME if the directory already exists we throw a PetalsException, it
        // is not correct except if we want to avoid the installation of a SU
        // that has been previously installed
        // TODO test if this is right to
        // test the existense of the
        // directory
        if (!installationRoot.exists() && !installationRoot.mkdirs()) {
            throw new PetalsException(
                    "Failed to create installation root dir: "
                            + installationRoot.getAbsolutePath());
        }
        // explode package Zip archive file into repository
        ZipFile zipArchive = null;
        try {
            File archiveFile = new File(suZipLocation);
            zipArchive = new ZipFile(archiveFile);
        } catch (ZipException ze) {
            throw new PetalsException("Unable to open Zip archive file: "
                    + suZipLocation.toString(), ze);
        } catch (IOException ioe) {
            throw new PetalsException("Unexpected exception.", ioe);
        }
        ZipUtil.explodeIntoDirectory(zipArchive, installationRoot);
        return installationRoot.toURI();
    }

    public String getComponentsDirectory() {
        return getRepositoryDirectory() + File.separator + COMPONENTS_PATH;
    }

    public String getServiceAssembliesDirectory() {
        return getRepositoryDirectory() + File.separator
                + SERVICEASSEMBLIES_PATH;
    }

    public String getSharedLibsDirectory() {
        return getRepositoryDirectory() + File.separator + SHAREDLIBS_PATH;
    }

    public String getLostPlusFoundDirectory() {
        return SystemUtil.getPetalsInstallDirectory().getAbsolutePath()
                + File.separator + LOSTPLUSFOUND_PATH;
    }

    public File getComponentInstallRoot(String componentId) {
        String path = getComponentRoot(componentId) + File.separator
                + INSTALL_PATH;
        return new File(path);
    }

    public File getComponentRoot(String componentId) {
        String path = getComponentsDirectory() + File.separator + componentId;
        return new File(path);
    }

    public File getComponentWorkRoot(String componentId) {
        String path = getComponentRoot(componentId) + File.separator
                + WORK_PATH;
        return new File(path);
    }

    public File getSharedLibRoot(String slId) {
        String path = getSharedLibsDirectory() + File.separator + slId;
        return new File(path);
    }

    public File getSharedLibWorkRoot(String slId) {
        String path = getSharedLibRoot(slId) + File.separator + WORK_PATH;
        return new File(path);
    }

    public File getSharedLibInstallRoot(String slId) {
        String path = getSharedLibRoot(slId) + File.separator + INSTALL_PATH;
        return new File(path);
    }

    public File getServiceAssemblyRoot(String saId) {
        String path = getServiceAssembliesDirectory() + File.separator + saId;
        return new File(path);
    }

    public File getServiceAssemblyWorkRoot(String saId) {
        String path = getServiceAssemblyRoot(saId) + File.separator + WORK_PATH;
        return new File(path);
    }

    public File getServiceAssemblyInstallRoot(String saId) {
        String path = getServiceAssemblyRoot(saId) + File.separator
                + INSTALL_PATH;
        return new File(path);
    }

    public String getFcState() {
        return STARTED.equals(fcState) ? STARTED : STOPPED;
    }

    public String[] listFc() {
        return new String[0];
    }

    public Object lookupFc(String clientItfName)
        throws NoSuchInterfaceException {
            throw new NoSuchInterfaceException(clientItfName);
    }

    /**
     * @see org.objectweb.petals.repository.RepositoryService#removePackage(java.lang.String)
     */
    protected boolean removePackage(String entityID, EntityType entityType) {
        boolean removed = true;

        File entityRoot = null;

        switch (entityType) {
        case COMPONENT_TYPE:
            entityRoot = getComponentRoot(entityID);
            break;
        case SA_TYPE:
            entityRoot = getServiceAssemblyRoot(entityID);
            break;
        case SL_TYPE:
            entityRoot = getSharedLibRoot(entityID);
            break;
        }

        Timer t = new Timer();
        RemovePackageTimerTask task = new RemovePackageTimerTask(t, entityRoot,
                log);
        t.schedule(task, 0, DELETION_DELAY * 1000);

        return removed;
    }

    public boolean removeComponentPackage(String componentId) {
        return removePackage(componentId, EntityType.COMPONENT_TYPE);
    }

    public boolean removeSharedLibPackage(String slId) {
        return removePackage(slId, EntityType.SL_TYPE);
    }

    public boolean removeServiceAssemblyPackage(String saId) {
        return removePackage(saId, EntityType.SA_TYPE);
    }

    public void startFc() throws IllegalLifeCycleException {
        init();

        fcState = STARTED;

    }

    public void stopFc() throws IllegalLifeCycleException {
        shutdown();

        fcState = STOPPED;

    }

    public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
        IllegalBindingException, IllegalLifeCycleException {
        if (LOGGER_NAME.equals(clientItfName)) {
            logger = null;
        } else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
            // nothing
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }

    }

    protected String getRepositoryDirectory() {
        if (repositoryDirectory == null) {
            repositoryDirectory = SystemUtil.getPetalsInstallDirectory()
                    .getAbsolutePath()
                    + File.separator + REPOSITORY_PATH;
        }
        return repositoryDirectory;
    }

    protected void init() {
        log = new LoggingUtil(logger);
        log.start();
        log.end();
    }

    protected void shutdown() {
        log.start();

        logger = null;

        log.end();
    }

    /**
     * TimerTask class used to remove directories when uninstalling an element.
     * A obscured problem with the JVM cache force us to try to delete the
     * directory a repeated time, until the cache is cleared. Sometimes this
     * cache is not cleared before the end of the task, and the directory is not
     * deleted. <br>
     * Note that the classloader of the element to uninstall tries to clean this
     * cache before deleting the directory.
     * 
     * @author alouis
     * 
     */
    class RemovePackageTimerTask extends TimerTask {
        private File dir;

        private LoggingUtil log;

        private Timer timer;

        private int tries;

        RemovePackageTimerTask(Timer timer, File directory, LoggingUtil log) {
            this.dir = directory;
            this.log = log;
            this.timer = timer;
        }

        @Override
        public void run() {
            if (dir.exists()) {
                try {
                    tries++;
                    FileUtils.deleteDirectory(dir);
                    log.debug("Repository deleted : " + dir);
                } catch (IOException e) {
                    log.warning("Failed to delete repository (" + tries + "/"
                            + DELETION_ATTEMPT + " attempts) " + dir, e);

                    if (tries > DELETION_ATTEMPT) {
                        log.error("Failed to delete repository : " + dir + "."
                                + e.getMessage(), e);
                        timer.cancel();
                    }
                }
            }
        }

    }
}